module Contender
  # Synchronization aid that allows one or more threads to wait until a set of operations being
  # performed in other threads complete
  class CountdownLatch
    # @return [Integer]
    attr_reader :count

    # @param [Integer] initial
    # @return [undefined]
    def initialize(initial)
      raise ArgumentError if initial < 0

      @count = initial

      @mutex = Mutex.new
      @condition = ConditionVariable.new
    end

    # Performs an ordered decrement operation for this latch
    #
    # @api public
    # @return [undefined]
    def countdown
      @mutex.synchronize do
        @count -= 1 if @count > 0
        @condition.broadcast if @count == 0
      end
    end

    # Waits until either the latch reaches zero or the timeout is reached
    #
    # @api public
    # @param [Float] timeout
    # @return [Boolean] True if the latch reached zero, false if the timeout was reached
    def await(timeout = nil)
      @mutex.synchronize do
        return true if @count == 0
        @condition.wait @mutex, timeout

        @count == 0
      end
    end
  end # CountdownLatch
end
