module Contender
  class FutureTask < Future
    # @param [Object] callable
    # @return [undefined]
    def initialize(callable)
      @callable = callable

      @mutex = Mutex.new
      @condition = ConditionVariable.new

      @state = :ready
    end

    # @api public
    # @param [Boolean] should_interrupt
    # @return [Boolean] True if this future was cancelled
    def cancel(should_interrupt)
      @mutex.synchronize do
        return false if done?

        if @state == :running
          return false unless should_interrupt
          @thread.raise InterruptError
        end

        @state = :cancelled
        @condition.broadcast
      end

      return true
    end

    # @api public
    # @return [Boolean]
    def cancelled?
      @state == :cancelled
    end

    # @api public
    # @return [Boolean]
    def done?
      @state == :ran || @state == :cancelled
    end

    # @api public
    # @raise [ExecutionError] If the result of the operation was an exception
    # @raise [TimeoutError] If the timeout was reached before the operation completed
    # @raise [CancellationError] If the operation was cancelled
    # @param [Integer] timeout Time to wait for the result
    # @return [Object] The result of the future
    def result(timeout = nil)
      @mutex.synchronize do
        unless done?
          @condition.wait @mutex, timeout

          unless done?
            raise TimeoutError, 'The operation did not complete before the given timeout'
          end
        end

        if @state == :cancelled
          raise CancellationError, 'Task was cancelled before it could be completed'
        end
      end

      if @exception
        raise ExecutionError, @exception
      end

      @result
    end

    # @api public
    # @return [undefined]
    def call
      @mutex.synchronize do
        return unless @state == :ready

        @state = :running
        @thread = Thread.current
      end

      begin
        @result = @callable.call
      rescue => exception
        @exception = exception
      end

      @mutex.synchronize do
        return unless @state == :running

        @state = :ran
        @condition.broadcast
      end
    ensure
      @thread = nil
    end
  end # FutureTask
end
