module Contender
  class LinkedQueue < Queue
    def initialize(capacity = FIXNUM_MAX)
      raise ArgumentError if capacity <= 0

      @capacity = capacity
      # Invariant: head.item == nil
      # Invariant: tail.next == nil
      @head = @tail = Node.new
      @count = Counter.new

      @take_lock = Monitor.new
      @not_empty = @take_lock.new_cond
      @put_lock = Monitor.new
      @not_full = @put_lock.new_cond
    end

    # @param [Object] element
    # @param [Float] timeout
    # @return [Boolean]
    def offer(element, timeout = nil)
      validate_element element

      count = 0
      node = Node.new element

      @put_lock.synchronize do
        if size == @capacity
          return false unless timeout

          @not_full.wait timeout
          return false if size == @capacity
        end

        enqueue node
        count = @count.increment
        if count < @capacity
          @not_full.signal
        end
      end

      if count > 0
        signal_not_empty
      end

      return true
    end

    # @param [Object] element
    # @return [undefined]
    def put(element)
      validate_element element

      count = 0
      node = Node.new element

      @put_lock.synchronize do
        while size == @capacity
          @not_full.wait
        end

        enqueue node
        count = @count.increment
        if count < @capacity
          @not_full.signal
        end
      end

      if count > 0
        signal_not_empty
      end
    end

    # @param [Float] timeout
    # @return [Object]
    def poll(timeout = nil)
      count = 0
      element = nil

      @take_lock.synchronize do
        if size == 0
          return unless timeout

          @not_empty.wait timeout
          return if size == 0
        end

        element = dequeue
        count = @count.decrement

        if count > 0
          @not_empty.signal
        end
      end

      if count < @capacity
        signal_not_full
      end

      return element
    end

    # @return [Object]
    def take
      count = 0
      element = nil

      @take_lock.synchronize do
        while size == 0
          @not_empty.wait
        end

        element = dequeue
        count = @count.decrement

        if count > 0
          @not_empty.signal
        end
      end

      if count < @capacity
        signal_not_full
      end

      return element
    end

    # @return [Object]
    def peek
      return if size == 0

      @take_lock.synchronize do
        first = @head.next
        return first.element if first
      end
    end

    # @return [undefined]
    def clear
      raise NotImplementedError
    end

    # @yield [Object]
    # @return [undefined]
    def each(&block)
      to_a.each &block
    end

    # @param [Object] element
    # @return [Boolean]
    def delete(element)
      return false if element.nil?

      full_synchronize do
        trail = @head
        interior = trail.next

        while interior
          if element == interior.element
            unlink interior, trail
            return true
          end

          trail = interior
          interior = interior.next
        end
      end

      return false
    end

    # @param [Array] target
    # @param [Integer] max_elements
    # @return [Integer]
    def drain_to(target, max_elements = FIXNUM_MAX)
      should_signal = false

      @take_lock.synchronize do
        n = [max_elements, size].min
        i = 0

        head = @head

        begin
          while i < n
            interior = head.next
            target.add interior.element!
            head.next = head
            head = interior

            i += 1
          end

          return n
        ensure
          if i > 0
            @head = head
            should_signal = @count.get_and_add(-i) == capacity
          end
        end
      end
    ensure
      signal_not_full if should_signal
    end

    def include?(element)
      return false if element.nil?

      full_synchronize do
        current = @head.next
        while current
          return true if element == current.element
          current = current.next
        end
      end

      return false
    end

    # @return [Integer]
    def capacity_remaining
      @capacity - @count.get
    end

    # @return [Integer]
    def size
      @count.get
    end

    # @return [Array]
    def to_a
      full_synchronize do
        result = Array.new size

        current = @head.next
        while current
          result.push current.element
          current = current.next
        end

        result
      end
    end

  protected

    # @yield
    # @return [undefined]
    def full_synchronize
      @take_lock.mon_enter
      @put_lock.mon_enter

      begin
        yield
      ensure
        @take_lock.mon_exit
        @put_lock.mon_exit
      end
    end

  private

    def validate_element(element)
      if element.nil?
        raise ArgumentError, 'This queue does not support nil elements'
      end
    end

    def signal_not_empty
      @take_lock.synchronize do
        @not_empty.signal
      end
    end

    def signal_not_full
      @put_lock.synchronize do
        @not_full.signal
      end
    end

    # Must have full lock
    def unlink(interior, trail)
      # Release the reference to the element
      interior.clear_element
      trail.next = interior.next

      if @tail == interior
        @tail = trail
      end

      if @count.decrement < @capacity
        @not_full.signal
      end
    end

    # Must have put lock
    def enqueue(node)
      @tail = @tail.next = node
    end

    # Must have take lock
    def dequeue
      head = @head
      first = head.next
      head.next = head # help GC

      @head = first

      first.element!
    end

    # Definition of a node in the linked list
    class Node
      # @return [Object]
      attr_accessor :element
      # @return [Node]
      attr_accessor :next

      def initialize(element = nil)
        @element = element
      end

      def element!
        @element.tap do
          @element = nil
        end
      end

      def clear_element
        @element = nil
      end

      def clear_next
        @next = nil
      end
    end # Node
  end # LinkedQueue
end
