module Contender
  module Pool
    # Represents the strategy to use when the pool executor rejects a task
    class RejectionPolicy
      # @abstract
      # @param [Object] task
      # @param [PoolExecutor] executor
      # @return [undefined]
      def on_rejection(task, executor)
        raise NotImplementedError
      end
    end

    # Rejection policy that executes the task in the calling thread
    class ExecutionByCallerPolicy < RejectionPolicy
      # @param [Object] task
      # @param [PoolExecutor] executor
      # @return [undefined]
      def on_rejection(task, executor)
        return if executor.shutdown?
        task.call
      end
    end

    # Rejection policy that simply raises an exception
    class AbortPolicy < RejectionPolicy
      # @param [Object] task
      # @param [PoolExecutor] executor
      # @return [undefined]
      def on_rejection(task, executor)
        raise TaskRejectionError, "Task #{task} rejected by thread pool executor"
      end
    end

    TaskRejectionError = Class.new RuntimeError

    # Rejection policy that silently discards a task
    class DiscardPolicy < RejectionPolicy
      # @param [Object] task
      # @param [PoolExecutor] executor
      # @return [undefined]
      def on_rejection(task, executor)
        # Just ignore the task
      end
    end

    # Rejection policy that discards the first task in the queue and then enqueues the
    # task
    class DiscardOldestPolicy < RejectionPolicy
      # @param [Object] task
      # @param [PoolExecutor] executor
      # @return [undefined]
      def on_rejection(task, executor)
        return if executor.shutdown?

        executor.queue.poll
        executor.execute task
      end
    end # DiscardOldestPolicy
  end # Pool
end
